/*
 * Licensed to the Apache Software Foundation (ASF) under one 
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY 
 * KIND, either express or implied.  See the License for the 
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.stratos.cloud.controller.domain.kubernetes;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.stratos.cloud.controller.util.CloudControllerUtil;
import org.apache.stratos.common.constants.StratosConstants;
import org.apache.stratos.kubernetes.client.KubernetesApiClient;
import org.apache.stratos.kubernetes.client.KubernetesConstants;
import org.apache.stratos.kubernetes.client.exceptions.KubernetesClientException;

/**
 * Holds information about a Kubernetes Cluster.
 */
public class KubernetesClusterContext implements Serializable {

    private static final long serialVersionUID = -802025758806195791L;

    private String kubernetesClusterId;
    private int upperPort;
    private int lowerPort;
    private String masterIp;
    private String masterPort;
    private String endpoint;
    private String namespace;
    private List<Integer> servicePortSequence;

    private AtomicLong serviceSeqNo;
    private AtomicLong podSeqNo;
    private transient KubernetesApiClient kubApi;
    public static final long MAX_POD_ID = 99999999999999L;
    public static final long MAX_SERVICE_ID = 99999999999999L;

    public KubernetesClusterContext(String id, String masterIp, String masterPort, int lowerPort, int upperPort,
                                    String endpoint, String namespace) throws KubernetesClientException {
        this.servicePortSequence = new ArrayList<>();
        serviceSeqNo = new AtomicLong(0);
        podSeqNo = new AtomicLong(0);

        this.lowerPort = lowerPort;
        this.upperPort = upperPort;
        // Generate the ports
        initializeServicePortSequence(lowerPort, upperPort);
        this.kubernetesClusterId = id;
        this.masterIp = masterIp;
        this.masterPort = masterPort;
        this.endpoint = endpoint;
        this.namespace = namespace;
        this.setKubApi(new KubernetesApiClient(prepareEndpoint(), namespace));
    }

    private String prepareEndpoint() {
        if (endpoint != null) {
            return endpoint;
        }
        URL endpointURL = null;
        try {
            if (this.masterPort != null) {
                endpointURL = new URL("http", this.masterIp, Integer.parseInt(masterPort), "");
            } else {
                endpointURL = new URL("http", this.masterIp, "");
            }
        } catch (MalformedURLException e) {
            throw new RuntimeException(e.getMessage());
        }
        return endpointURL.toString();
    }

    public String getKubernetesClusterId() {
        return kubernetesClusterId;
    }

    public void setKubernetesClusterId(String kubernetesClusterId) {
        this.kubernetesClusterId = kubernetesClusterId;
    }

    public List<Integer> getServicePorts() {
        return servicePortSequence;
    }

    public void setServicePorts(List<Integer> servicePorts) {
        this.servicePortSequence = servicePorts;
    }

    /***
     * Get next available service port.
     *
     * @return
     */
    public int getNextServicePort() {
        if (servicePortSequence.isEmpty()) {
            return -1;
        }
        return servicePortSequence.remove(0);
    }

    /**
     * Deallocate a service port by adding it again to the sequence.
     *
     * @param port
     */
    public void deallocatePort(int port) {
        if (!servicePortSequence.contains(port)) {
            servicePortSequence.add(port);
            Collections.sort(servicePortSequence);
        }
    }

    /**
     * Initialize service port sequence according to the given port range.
     *
     * @param lowerPort
     * @param upperPort
     */
    private void initializeServicePortSequence(int lowerPort, int upperPort) {
        for (int port = lowerPort; port <= upperPort; port++) {
            servicePortSequence.add(port);
        }
    }

    public String getMasterIp() {
        return masterIp;
    }

    public void setMasterIp(String masterIp) {
        this.masterIp = masterIp;
    }

    public KubernetesApiClient getKubApi() throws KubernetesClientException {
        if (kubApi == null) {
            kubApi = new KubernetesApiClient(prepareEndpoint(), namespace);
        }
        return kubApi;
    }

    public void setKubApi(KubernetesApiClient kubApi) {
        this.kubApi = kubApi;
    }

    public int getUpperPort() {
        return upperPort;
    }

    public void setUpperPort(int upperPort) {
        this.upperPort = upperPort;
    }

    public int getLowerPort() {
        return lowerPort;
    }

    public void setLowerPort(int lowerPort) {
        this.lowerPort = lowerPort;
    }

    public long getNextServiceSeqNo() {
        // reset before we hit the max character length for Kub service id
        if (serviceSeqNo.get() > MAX_SERVICE_ID) {
            serviceSeqNo.set(0);
        }
        return serviceSeqNo.incrementAndGet();
    }

    public long getNextPodSeqNo() {
        // reset before we hit the max character length for Kub pod id
        if (podSeqNo.get() > MAX_POD_ID) {
            podSeqNo.set(0);
        }
        return podSeqNo.incrementAndGet();
    }

    public void updateKubClusterContextParams(KubernetesCluster kubernetesCluster) throws KubernetesClientException {
        this.masterIp = kubernetesCluster.getKubernetesMaster().getPrivateIPAddress();
        this.masterPort = CloudControllerUtil
                .getProperty(kubernetesCluster.getKubernetesMaster().getProperties(),
                        StratosConstants.KUBERNETES_MASTER_PORT);
        this.namespace = CloudControllerUtil.getProperty(kubernetesCluster.getKubernetesMaster().getProperties(),
                KubernetesConstants.KUBERNETES_NAMESPACE, KubernetesConstants.DEFAULT_NAMESPACE);
        this.endpoint = kubernetesCluster.getKubernetesMaster().getEndpoint();
        this.kubApi = new KubernetesApiClient(prepareEndpoint(), this.namespace);
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((servicePortSequence == null) ? 0 : servicePortSequence.hashCode());
        result = prime * result + ((kubernetesClusterId == null) ? 0 : kubernetesClusterId.hashCode());
        result = prime * result + lowerPort;
        result = prime * result + ((masterIp == null) ? 0 : masterIp.hashCode());
        result = prime * result + ((masterPort == null) ? 0 : masterPort.hashCode());
        result = prime * result + upperPort;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        KubernetesClusterContext other = (KubernetesClusterContext) obj;
        if (servicePortSequence == null) {
            if (other.servicePortSequence != null) {
                return false;
            }
        } else if (!servicePortSequence.equals(other.servicePortSequence)) {
            return false;
        }
        if (kubernetesClusterId == null) {
            if (other.kubernetesClusterId != null) {
                return false;
            }
        } else if (!kubernetesClusterId.equals(other.kubernetesClusterId)) {
            return false;
        }
        if (lowerPort != other.lowerPort) {
            return false;
        }
        if (masterIp == null) {
            if (other.masterIp != null) {
                return false;
            }
        } else if (!masterIp.equals(other.masterIp)) {
            return false;
        }
        if (masterPort == null) {
            if (other.masterPort != null) {
                return false;
            }
        } else if (!masterPort.equals(other.masterPort)) {
            return false;
        }
        if (endpoint == null) {
            if (other.endpoint != null) {
                return false;
            }
        } else if (!endpoint.equals(other.endpoint)) {
            return false;
        }
        if (upperPort != other.upperPort) {
            return false;
        }
        return true;
    }
}